# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Top of atmosphere incident solar radiation.

This code was inspired by third_party/py/pysolar, however that library is
intended as a surface of Earth solar radiation model that includes atmospheric
scattering effects. Of particularl note, the constants used to compute solar
irradiance are different. References included below.

Another major difference is that this module represents time using radians,
which greatly reduces the number of conversions by factors of pi.
"""

from __future__ import annotations

import datetime

from dinosaur import coordinate_systems
from dinosaur import primitive_equations
from dinosaur import scales
from dinosaur import typing
import jax
import jax.numpy as jnp
import numpy as np
import tree_math

units = scales.units

Array = typing.Array
Numeric = typing.Numeric

DAYS_PER_YEAR = 365.25
MINUTES_PER_DAY = 1440
SECONDS_PER_DAY = 86400
# TSI: Energy input to the top of the Earth's atmosphere
# https://www.ncei.noaa.gov/products/climate-data-records/total-solar-irradiance
TOTAL_SOLAR_IRRADIANCE = 1361 * units.W / units.meter**2
# Seasonal variation in apparent solar irradiance due to Earth-Sun distance
SOLAR_IRRADIANCE_VARIATION = 47 * units.W / units.meter**2  # .5 * 6.9% * TSI
# Approximate perihelion, when Earth is closest to the sun (Jan 3rd)
PERIHELION = 3 * 2 * jnp.pi / DAYS_PER_YEAR  # radians
# Approximate equinox (March 20 on non-leap year), when dihedral is zero
SPRING_EQUINOX = 79 * 2 * jnp.pi / DAYS_PER_YEAR  # radians
# Angle between Earth's rotational axis and its orbital axis
EARTH_AXIS_INCLINATION = 23.45 * jnp.pi / 180  # radians

# Reference datetime for WeatherBench
WB_REFERENCE_DATETIME = datetime.datetime(1979, 1, 1, 0, 0)


@tree_math.struct
class OrbitalTime:
  """Nondimensional time based on orbital dynamics.

  Attributes:
    orbital_phase: phase of the Earth's orbit around the Sun in radians. The
      values 0, 2pi correspond to January 1st, midnight UTC.
    synodic_phase: phase of the Earth's rotation around its axis in radians,
      relative to the Sun. The values 0, 2pi correspond to midnight UTC.
  """

  orbital_phase: float
  synodic_phase: float


def datetime64_to_datetime(when: np.datetime64) -> datetime.datetime:
  """Returns datetime corresponding to the provided numpy datetime64."""
  ts = (when - np.datetime64('1970-01-01T00:00:00')) / np.timedelta64(1, 's')
  return datetime.datetime.utcfromtimestamp(ts)


def days_in_year(when: datetime.datetime) -> int:
  """Returns the number of days in the year of the provided datetime."""
  return datetime.datetime(year=when.year, month=12, day=31).timetuple().tm_yday


def datetime_to_orbital_time(when: datetime.datetime) -> OrbitalTime:
  """Returns the OrbitalTime associated with the provided datetime."""
  days_this_year = days_in_year(when)
  full_days = when.timetuple().tm_yday - 1
  fraction_of_day = (60 * when.hour + when.minute) / MINUTES_PER_DAY
  fraction_of_year = (full_days + fraction_of_day) / days_this_year
  return OrbitalTime(
      orbital_phase=2 * jnp.pi * fraction_of_year,
      synodic_phase=2 * jnp.pi * fraction_of_day,
  )


def datetime_to_time(
    when: datetime.datetime | np.datetime64,
    physics_specs: primitive_equations.PrimitiveEquationsSpecs,
    reference_datetime: datetime.datetime | np.datetime64,
) -> float:
  """Returns nondimensional time corresponding to the specified datetime.

  Args:
    when: datetime for which to compute nondimensional time.
    physics_specs: object holding physical constants and definition of custom
      units to use for initialization of the state.
    reference_datetime: datetime corresponding to nondimensionalized time = 0.
  """
  if isinstance(when, np.datetime64):
    when = datetime64_to_datetime(when)
  if isinstance(reference_datetime, np.datetime64):
    reference_datetime = datetime64_to_datetime(reference_datetime)
  difference = when - reference_datetime
  days = difference.days + difference.seconds / SECONDS_PER_DAY
  return physics_specs.nondimensionalize(days * units.day)  # pytype: disable=bad-return-type  # jax-ndarray


def get_direct_solar_irradiance(
    orbital_phase: Numeric,
    mean_irradiance: Numeric = TOTAL_SOLAR_IRRADIANCE,
    variation: Numeric = SOLAR_IRRADIANCE_VARIATION,
    perihelion: Numeric = PERIHELION,
) -> jnp.ndarray:
  """Returns solar radiation flux incident on the top of the atmosphere.

  Formula includes 6.9% seasonal variation due to Earth's elliptical orbit, but
  neglects the 0.1% variation of the 11-year solar cycle (Schwabe cycle).
  https://earth.gsfc.nasa.gov/climate/research/solar-radiation
  https://en.wikipedia.org/wiki/Solar_constant

  Note that the default values here are different from third_party/py/pysolar,
  which uses `1160 + (75 * math.sin(2 * math.pi / 365 * (day - 275)))`. This is
  presumably because pysolar is intended for modeling radiation incident on
  Earth's surface, including atmospheric scattering.

  Args:
    orbital_phase: phase of the Earth's orbit around the Sun in radians. The
      values 0, 2pi correspond to January 1st, midnight UTC.
    mean_irradiance: average annual solar flux just outside Earth's atmosphere.
      Default is the "total solar irradiance", or "solar consant", in W/m^2.
    variation: amplitude of fluctuation in solar flux due to Earth's elliptical
      orbit. Default is 0.5 * 6.9% of TSI. Units should match mean_irradiance.
    perihelion: orbital phase in radians where the Earth is closest to the Sun.
  """
  return mean_irradiance + variation * jnp.cos(orbital_phase - perihelion)


def get_declination(orbital_phase: Numeric) -> jnp.ndarray:
  """Returns angle between the Earth-Sun line and the Earth equitorial plane."""
  # https://en.wikipedia.org/wiki/Declination
  return EARTH_AXIS_INCLINATION * jnp.sin(orbital_phase - SPRING_EQUINOX)


def equation_of_time(orbital_phase: Numeric) -> jnp.ndarray:
  """Returns the value to add to mean solar time to get actual solar time."""
  # https://en.wikipedia.org/wiki/Equation_of_time
  b = orbital_phase - SPRING_EQUINOX
  added_minutes = 9.87 * jnp.sin(2 * b) - 7.53 * jnp.cos(b) - 1.5 * jnp.sin(b)
  # Output normalized as a correction to synodic_phase
  return 2 * jnp.pi * added_minutes / MINUTES_PER_DAY


def get_hour_angle(
    orbital_phase: Numeric, synodic_phase: Numeric, longitude: Array
) -> jnp.ndarray:
  """Angular displacement of the sun east or west of the local meridian."""
  # https://en.wikipedia.org/wiki/Hour_angle
  solar_time = synodic_phase + equation_of_time(orbital_phase) + longitude
  return solar_time - jnp.pi


def get_solar_sin_altitude(
    orbital_phase: Numeric,
    synodic_phase: Numeric,
    longitude: Array,
    latitude: Array,
) -> jnp.ndarray:
  """Returns sine of the solar altitude angle."""
  # https://en.wikipedia.org/wiki/Solar_zenith_angle
  declination = get_declination(orbital_phase)
  hour_angle = get_hour_angle(orbital_phase, synodic_phase, longitude)
  first_term = jnp.cos(latitude) * jnp.cos(declination) * jnp.cos(hour_angle)
  second_term = jnp.sin(latitude) * jnp.sin(declination)
  return first_term + second_term


def get_radiation_flux(
    orbital_time: OrbitalTime,
    longitude: Array,
    latitude: Array,
    mean_irradiance: Numeric = TOTAL_SOLAR_IRRADIANCE,
    variation: Numeric = SOLAR_IRRADIANCE_VARIATION,
) -> jnp.ndarray:
  """Returns TOA incident radiation flux."""
  sin_altitude = get_solar_sin_altitude(
      orbital_phase=orbital_time.orbital_phase,
      synodic_phase=orbital_time.synodic_phase,
      longitude=longitude,
      latitude=latitude,
  )
  is_daytime = sin_altitude > 0
  flux = get_direct_solar_irradiance(
      orbital_phase=orbital_time.orbital_phase,
      mean_irradiance=mean_irradiance,
      variation=variation,
  )
  return flux * is_daytime * sin_altitude


def get_normalized_radiation_flux(
    orbital_time: OrbitalTime,
    longitude: Array,
    latitude: Array,
    mean_irradiance: Numeric = TOTAL_SOLAR_IRRADIANCE,
    variation: Numeric = SOLAR_IRRADIANCE_VARIATION,
) -> jnp.ndarray:
  """Returns TOA incident radiation flux, normalized to between -1 and 1."""
  scale = mean_irradiance + variation
  return get_radiation_flux(
      orbital_time,
      longitude,
      latitude,
      mean_irradiance=mean_irradiance / scale,
      variation=variation / scale,
  )


class SolarRadiation:
  """Top of atmosphere incident solar radiation (TISR)."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      physics_specs: primitive_equations.PrimitiveEquationsSpecs,
      reference_datetime: datetime.datetime | np.datetime64,
  ):
    """Initialize SolarRadiation.

    Args:
      coords: horizontal and vertical descritization.
      physics_specs: object holding physical constants and definition of custom
        units to use for initialization of the state.
      reference_datetime: datetime corresponding to nondimensionalized time = 0.
    """
    if isinstance(reference_datetime, np.datetime64):
      reference_datetime = datetime64_to_datetime(reference_datetime)
    self.reference_datetime = reference_datetime
    self.reference_orbital_time = datetime_to_orbital_time(reference_datetime)

    self.coords = coords
    self.lon, sin_lat = self.coords.horizontal.nodal_mesh
    self.lat = np.arcsin(sin_lat)
    self.orbital_rate = jax.tree_util.tree_map(
        physics_specs.nondimensionalize,
        OrbitalTime(2 * jnp.pi / units.year, 2 * jnp.pi / units.day),
    )

    self.total_solar_irradiance = physics_specs.nondimensionalize(
        TOTAL_SOLAR_IRRADIANCE
    )
    self.solar_irradiance_variation = physics_specs.nondimensionalize(
        SOLAR_IRRADIANCE_VARIATION
    )
    self.physics_specs = physics_specs

  def datetime_to_time(self, when: datetime.datetime | np.datetime64) -> float:
    """Returns nondimensional time corresponding to the specified datetime."""
    return datetime_to_time(when, self.physics_specs, self.reference_datetime)

  def time_to_orbital_time(self, time: Numeric) -> OrbitalTime:
    """Returns the OribtalTime corresponding to the specified nondim time."""
    orbital_time = self.reference_orbital_time + self.orbital_rate * time
    # Reduce the magnitude of the result to avoid loss of precision errors
    # downstream. Avoid jnp.fmod, which is not very precise on float32.
    orbital_time -= orbital_time // (2 * jnp.pi) * (2 * jnp.pi)
    return orbital_time

  def solar_hour_angle(self, time: Numeric) -> jnp.ndarray:
    """Returns solar hour angle in radians."""
    now = self.time_to_orbital_time(time)
    return get_hour_angle(
        orbital_phase=now.orbital_phase,
        synodic_phase=now.synodic_phase,
        longitude=self.lon,
    )

  def radiation_flux(self, time: Numeric) -> jnp.ndarray:
    """Returns non-dimensionalized TOA incident solar radiation flux."""
    now = self.time_to_orbital_time(time)
    return get_radiation_flux(
        now,
        self.lon,
        self.lat,
        mean_irradiance=self.total_solar_irradiance,
        variation=self.solar_irradiance_variation,
    )

  @classmethod
  def normalized(
      cls,
      coords: coordinate_systems.CoordinateSystem,
      physics_specs: primitive_equations.PrimitiveEquationsSpecs,
      reference_datetime: datetime.datetime | np.datetime64,
  ) -> SolarRadiation:
    """Initialize SolarRadiation for normalized solar radiation."""
    this = cls(
        coords=coords,
        physics_specs=physics_specs,
        reference_datetime=reference_datetime,
    )
    scale = this.total_solar_irradiance + this.solar_irradiance_variation
    this.total_solar_irradiance /= scale
    this.solar_irradiance_variation /= scale
    return this



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for jax_numpy_utils."""
import math

from absl.testing import absltest
from absl.testing import parameterized
import chex

from dinosaur import jax_numpy_utils

import jax
import numpy as np


chex.set_n_cpu_devices(8)  # set at top level so it works in pytest


class CumsumTest(parameterized.TestCase):

  @parameterized.parameters({'axis': 0}, {'axis': 1}, {'axis': -1})
  def test_cumsum(self, axis):
    x = np.random.RandomState(0).randn(3, 4)
    expected = np.cumsum(x, axis=axis)
    actual = jax_numpy_utils.cumsum(x, axis=axis, method='dot')
    np.testing.assert_allclose(actual, expected, atol=1e-6)

  @parameterized.parameters({'axis': 0}, {'axis': 1}, {'axis': -1})
  def test_reverse_cumsum_consistency(self, axis):
    x = np.random.RandomState(0).randn(3, 4)
    expected = jax_numpy_utils.reverse_cumsum(x, axis=axis, method='jax')
    actual = jax_numpy_utils.reverse_cumsum(x, axis=axis, method='dot')
    np.testing.assert_allclose(actual, expected, atol=1e-6)

  def test_shmap(self):
    x = np.random.RandomState(0).randn(16)
    sharding = jax.sharding.NamedSharding(
        jax.sharding.Mesh(jax.devices(), axis_names=['z']),
        jax.sharding.PartitionSpec('z'),
    )
    y = jax.device_put(x, sharding)

    expected = jax_numpy_utils.cumsum(x, axis=0)
    actual = jax_numpy_utils.cumsum(y, axis=0, sharding=sharding)
    np.testing.assert_allclose(actual, expected, atol=1e-6)

    expected = jax_numpy_utils.reverse_cumsum(x, axis=0)
    actual = jax_numpy_utils.reverse_cumsum(y, axis=0, sharding=sharding)
    np.testing.assert_allclose(actual, expected, atol=1e-6)

  @parameterized.parameters({'axis': 0}, {'axis': 1}, {'axis': -1})
  def test_shmap_2d(self, axis):
    x = np.random.RandomState(0).randn(16, 3)
    sharding = jax.sharding.NamedSharding(
        jax.sharding.Mesh(jax.devices(), axis_names=['z']),
        jax.sharding.PartitionSpec('z', None),
    )
    y = jax.device_put(x, sharding)

    expected = jax_numpy_utils.cumsum(x, axis=axis)
    actual = jax_numpy_utils.cumsum(y, axis=axis, sharding=sharding)
    np.testing.assert_allclose(actual, expected, atol=1e-6)

    expected = jax_numpy_utils.reverse_cumsum(x, axis=axis)
    actual = jax_numpy_utils.reverse_cumsum(y, axis=axis, sharding=sharding)
    np.testing.assert_allclose(actual, expected, atol=1e-6)


class ShiftTest(parameterized.TestCase):

  @parameterized.parameters(
      {
          'inputs': np.array([1, 2, 3]),
          'offset': 0,
          'expected': np.array([1, 2, 3]),
      },
      {
          'inputs': np.array([1, 2, 3]),
          'offset': +1,
          'expected': np.array([0, 1, 2]),
      },
      {
          'inputs': np.array([1, 2, 3]),
          'offset': -1,
          'expected': np.array([2, 3, 0]),
      },
      {
          'inputs': np.array([1, 2, 3]),
          'offset': +5,
          'expected': np.array([0, 0, 0]),
      },
      {
          'inputs': np.array([1, 2, 3]),
          'offset': -5,
          'expected': np.array([0, 0, 0]),
      },
      {
          'inputs': np.ones((2, 3)),
          'offset': +1,
          'axis': 0,
          'expected': np.stack([np.zeros(3), np.ones(3)]),
      },
      {
          'inputs': np.ones((2, 3)),
          'offset': -1,
          'axis': -2,
          'expected': np.stack([np.ones(3), np.zeros(3)]),
      },
      {
          'inputs': np.ones((2, 3), int),
          'offset': +1,
          'axis': 1,
          'expected': np.array([[0, 1, 1], [0, 1, 1]]),
      },
  )
  def test_shift(self, inputs, offset, expected, axis=0):
    actual = jax_numpy_utils.shift(inputs, offset, axis)
    np.testing.assert_array_equal(expected, actual)


P = jax.sharding.PartitionSpec


_EINSUM_CASES = [
    {
        'subscripts': 'ij,jk->ik',
        'lhs_shape': (2, 2),
        'rhs_shape': (2, 2),
        'axis_names': ['x', 'y'],
        'mesh_shape': (2, 2),
        'rhs_spec': P('x', 'y'),
        'out_spec': P('x', 'y'),
    },
    {
        'subscripts': 'ij,jk->ik',
        'lhs_shape': (4, 4),
        'rhs_shape': (4, 3),
        'axis_names': ['x', 'y'],
        'mesh_shape': (4, 1),
        'rhs_spec': P('x', 'y'),
        'out_spec': P('x', 'y'),
    },
    {
        'subscripts': 'ij,jk->ik',
        'lhs_shape': (3, 3),
        'rhs_shape': (3, 4),
        'axis_names': ['x', 'y'],
        'mesh_shape': (1, 4),
        'rhs_spec': P('x', 'y'),
        'out_spec': P('x', 'y'),
    },
    {
        # inverse Legendre
        'subscripts': 'mjl,zsml->zsmj',
        'lhs_shape': (2, 2, 2),
        'rhs_shape': (3, 2, 2, 2),
        'axis_names': ['z', 'x', 'y'],
        'mesh_shape': (1, 2, 2),
        'rhs_spec': P('z', None, 'x', 'y'),
        'out_spec': P('z', None, 'x', 'y'),
    },
    {
        # inverse Fourier
        'subscripts': 'ism,zsmj->zij',
        'lhs_shape': (2, 2, 2),
        'rhs_shape': (3, 2, 2, 2),
        'axis_names': ['z', 'x', 'y'],
        'mesh_shape': (1, 2, 2),
        'rhs_spec': P('z', None, 'x', 'y'),
        'out_spec': P('z', 'x', 'y'),
    },
    {
        # forward Fourier
        'subscripts': 'ism,zij->zsmj',
        'lhs_shape': (4, 2, 4),
        'rhs_shape': (3, 4, 2),
        'axis_names': ['z', 'x', 'y'],
        'mesh_shape': (1, 4, 2),
        'rhs_spec': P('z', 'x', 'y'),
        'out_spec': P('z', None, 'x', 'y'),
    },
    {
        # forward Legendre
        'subscripts': 'mjl,zsmj->zsml',
        'lhs_shape': (2, 2, 2),
        'rhs_shape': (4, 2, 2, 2),
        'axis_names': ['z', 'x', 'y'],
        'mesh_shape': (2, 2, 2),
        'rhs_spec': P('z', None, 'x', 'y'),
        'out_spec': P('z', None, 'x', 'y'),
    },
    {
        # vertical matvec
        'subscripts': 'gh,hml->gml',
        'lhs_shape': (8, 8),
        'rhs_shape': (8, 1, 1),
        'axis_names': ['z', 'x', 'y'],
        'mesh_shape': (8, 1, 1),
        'rhs_spec': P('z', 'x', 'y'),
        'out_spec': P('z', 'x', 'y'),
    },
    {
        # vertical matvec per wavenumber
        'subscripts': 'lgh,hml->gml',
        'lhs_shape': (2, 4, 4),
        'rhs_shape': (4, 2, 2),
        'axis_names': ['z', 'x', 'y'],
        'mesh_shape': (4, 1, 2),
        'rhs_spec': P('z', 'x', 'y'),
        'out_spec': P('z', 'x', 'y'),
    },
]
_GATHER_CASES = [dict(case, gather_inputs=True) for case in _EINSUM_CASES]
_SCATTER_CASES = [dict(case, gather_inputs=False) for case in _EINSUM_CASES]
_FLEX_CASES = [dict(case, gather_inputs=None) for case in _EINSUM_CASES]


class ShardedEinsum(parameterized.TestCase):

  @parameterized.parameters(*_GATHER_CASES, *_SCATTER_CASES, *_FLEX_CASES)
  def test_sharded_einsum(
      self, subscripts, lhs_shape, rhs_shape, axis_names, mesh_shape, **kwargs
  ):
    mesh_size = math.prod(mesh_shape)
    devices = np.array(jax.devices()[:mesh_size]).reshape(mesh_shape)
    mesh = jax.sharding.Mesh(devices, axis_names)
    lhs = 1.0 * np.arange(math.prod(lhs_shape)).reshape(lhs_shape)
    rhs = 2.0 * np.arange(math.prod(rhs_shape)).reshape(rhs_shape)
    expected = np.einsum(subscripts, lhs, rhs)
    actual = jax_numpy_utils.sharded_einsum(
        subscripts, lhs, rhs, mesh=mesh, **kwargs
    )
    np.testing.assert_allclose(actual, expected, atol=1e-6)


if __name__ == '__main__':
  jax.config.update('jax_traceback_filtering', 'off')
  absltest.main()



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for associated_legendre."""

from absl.testing import absltest
from absl.testing import parameterized

from dinosaur import associated_legendre

import numpy as np
import scipy.special as sps


class AssociatedLegendreTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(n=4),
      dict(n=11),
      dict(n=32),
  )
  def testOrthonormality(self, n):
    """Tests that the computed associated Legendre functions are orthonormal."""
    z, w = associated_legendre.gauss_legendre_nodes(n)
    p = associated_legendre.evaluate(n, n, z)
    i = np.eye(n, dtype=np.float64)
    inner_products = np.einsum('mil,mik,i->mlk', p, p, w)
    for m in range(n):
      np.testing.assert_allclose(i, inner_products[m], atol=1e-8)
      i[m, m] = 0

  @parameterized.parameters(
      dict(m=4, l=4),
      dict(m=3, l=8),
      dict(m=12, l=20)
  )
  def testAgainstScipy(self, m, l):
    x, _ = associated_legendre.gauss_legendre_nodes(l + 1)
    p_lm = associated_legendre.evaluate(m + 1, l + 1, x)[-1, :, -1]
    q_lm = sps.lpmv(m, l, x)

    # We assert that `q_lm = ratio * p_lm` for some scalar `ratio`.
    ratio = q_lm[0] / p_lm[0]
    for p_j, q_j in zip(p_lm, q_lm):
      np.testing.assert_almost_equal(q_j / ratio, p_j)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines `Grid` that holds vertical and horizontal grids."""

from __future__ import annotations

import dataclasses
from typing import Any, Callable, Sequence

from dinosaur import layer_coordinates
from dinosaur import pytree_utils
from dinosaur import sigma_coordinates
from dinosaur import spherical_harmonic
from dinosaur import typing

import jax
import jax.numpy as jnp
import numpy as np

# keys that correspond to the grid/coordinate type in asdict.
HORIZONTAL_COORD_TYPE_KEY = 'horizontal_grid_type'
VERTICAL_COORD_TYPE_KEY = 'vertical_grid_type'
HorizontalGridTypes = spherical_harmonic.Grid
VerticalCoordinateTypes = (
    layer_coordinates.LayerCoordinates
    | sigma_coordinates.SigmaCoordinates
    | Any
)


P = jax.sharding.PartitionSpec


def _with_sharding_constraint(
    x: typing.Pytree,
    sharding: jax.sharding.NamedSharding | None,
) -> typing.Pytree:
  """Ensure a sharing constraint on all non-scalar arrays in a pytree."""
  if sharding is None:
    return x  # unsharded

  if len(sharding.spec) != 3:
    raise ValueError(f'partition spec does not have length 3: {sharding.spec}')

  def f(y: jax.Array) -> jax.Array:
    if y.ndim == 1 and y.dtype == jnp.uint32:
      return y  # prng key

    if y.ndim not in {2, 3}:
      raise ValueError(f'can only shard 2D or 3D arrays: {y.shape=}')

    # TODO(shoyer): refactor our ML codebase so we only need to shard 3D arrays.
    # This currently is only needed for stochastic perturbations.
    if y.ndim == 2:
      spec = P(*sharding.spec[1:])
      sharding_ = jax.sharding.NamedSharding(sharding.mesh, spec)
    elif y.shape[0] == 1:
      # surface level variable
      spec = P(None, *sharding.spec[1:])
      sharding_ = jax.sharding.NamedSharding(sharding.mesh, spec)
    else:
      sharding_ = sharding
    return jax.lax.with_sharding_constraint(y, sharding_)

  try:
    return pytree_utils.tree_map_over_nonscalars(f, x)
  except ValueError as e:
    shapes = jax.tree_util.tree_map(jnp.shape, x)
    raise ValueError(f'failed to shard pytree with shapes {shapes}') from e


@dataclasses.dataclass(frozen=True)
class CoordinateSystem:
  """Contains horizontal and vertical grid data.

  Attributes:
    horizontal: object describing discretization of the horizontal plane.
    vertical: object describing discretization of the vertical coordinate.
    spmd_mesh: mesh to use for parallelism in the single program multiple device
      (SPMD) paradigm with distributed JAX arrays, if any. Required if using
      model parallelism.
    physics_partition_spec: partition spec for "physics" calculations that act
      only in the vertical direction (e.g., ML corrections and vertical
      interpolation). We also use this partitioning for loading data from disk
      and calculating losses.
    dycore_partition_spec: partition spec for "dycore" calculations across all
      spatial dimensions (e.g., spherical harmonic transforms and solving the
      primitive equations).
  """

  horizontal: HorizontalGridTypes
  vertical: VerticalCoordinateTypes
  spmd_mesh: jax.sharding.Mesh | None = None

  # In principle, these partition spec arguments can be customized, but it is
  # not recommended.

  # The logical dimension names on `spmd_mesh` correspond to those used in the
  # dycore.
  dycore_partition_spec: jax.sharding.PartitionSpec = P('z', 'x', 'y')

  # For cases where we don't want to use vertical partitioning (e.g., because
  # we are applying a column-wise neural net), we merge 'z' into one of the
  # spatial dimensions instead. In principle, it shouldn't matter whether we
  # merge 'z' into 'x' or 'y', but in practice XLA SPMD generates more efficient
  # code (avoiding "full rematerialization") if we merge into the adjacent
  # dimension. Both ('x', 'z') and ('z', 'x') work, but training with ('x', 'z')
  # on TL255 is slightly faster (0.4% overall).
  physics_partition_spec: jax.sharding.PartitionSpec = P(None, ('x', 'z'), 'y')

  def __post_init__(self):
    if self.spmd_mesh is not None:
      if not {'x', 'y', 'z'} <= set(self.spmd_mesh.axis_names):
        raise ValueError(
            "mesh is missing one or more of the required axis names 'x', 'y' "
            f"and 'z': {self.spmd_mesh}"
        )
    # ensure a consistent mesh
    horizontal = dataclasses.replace(self.horizontal, spmd_mesh=self.spmd_mesh)
    object.__setattr__(self, 'horizontal', horizontal)

  def _get_sharding(
      self, partition_spec: jax.sharding.PartitionSpec
  ) -> jax.sharding.NamedSharding | None:
    if self.spmd_mesh is None:
      return None
    return jax.sharding.NamedSharding(self.spmd_mesh, partition_spec)

  @property
  def physics_sharding(self) -> jax.sharding.NamedSharding | None:
    """How to shard arrays for "physics" calculations."""
    return self._get_sharding(self.physics_partition_spec)

  def with_physics_sharding(self, x: typing.PyTreeState) -> typing.PyTreeState:
    """Enforce a "physics" sharding constraint on a pytree."""
    return _with_sharding_constraint(x, self.physics_sharding)

  @property
  def dycore_sharding(self) -> jax.sharding.NamedSharding | None:
    """How to shard arrays for nodal or modal dycore calculations."""
    return self._get_sharding(self.dycore_partition_spec)

  def with_dycore_sharding(self, x: typing.PyTreeState) -> typing.PyTreeState:
    """Enforce a dycore sharding constraint on a pytree."""
    return _with_sharding_constraint(x, self.dycore_sharding)

  def dycore_to_physics_sharding(
      self, x: typing.PyTreeState
  ) -> typing.PyTreeState:
    """Transpose from dycore to physics sharding."""
    return self.with_physics_sharding(self.with_dycore_sharding(x))

  def physics_to_dycore_sharding(
      self, x: typing.PyTreeState
  ) -> typing.PyTreeState:
    """Transpose from physics to dycore sharding."""
    return self.with_dycore_sharding(self.with_physics_sharding(x))

  def asdict(self) -> ...:
    horizontal_keys = set(self.horizontal.asdict().keys())
    vertical_keys = set(self.vertical.asdict().keys())
    if horizontal_keys.intersection(vertical_keys):
      raise ValueError('keys in horizontal and vertical grids collide.')
    out = {**self.horizontal.asdict(), **self.vertical.asdict()}
    out[HORIZONTAL_COORD_TYPE_KEY] = type(self.horizontal).__name__
    out[VERTICAL_COORD_TYPE_KEY] = type(self.vertical).__name__
    return out

  @property
  def nodal_shape(self) -> tuple[int, int, int]:
    """Returns 3d nodal grid shape, vertical.shape + horizontal.shape."""
    return (self.vertical.layers,) + self.horizontal.nodal_shape

  @property
  def modal_shape(self) -> tuple[int, int, int]:
    """Returns 3d modal grid shape, vertical.shape + horizontal.shape."""
    return (self.vertical.layers,) + self.horizontal.modal_shape

  @property
  def surface_nodal_shape(self) -> tuple[int, int, int]:
    """Returns surface nodal grid shape, (1,) + horizontal.shape."""
    return (1,) + self.horizontal.nodal_shape

  @property
  def surface_modal_shape(self) -> tuple[int, int, int]:
    """Returns surface modal grid shape, (1,) + horizontal.shape."""
    return (1,) + self.horizontal.modal_shape


def get_spectral_downsample_fn(
    coords: CoordinateSystem,
    save_coords: CoordinateSystem,
    expect_same_vertical: bool = True,
) -> Callable[[typing.PyTreeState], typing.PyTreeState]:
  """Returns function that downsample state from `coords` to `save_coords`."""
  if expect_same_vertical and (coords.vertical != save_coords.vertical):
    raise ValueError('downsampling vertical resolution is not supported.')
  lon_wavenumber_slice = slice(0, save_coords.horizontal.modal_shape[0])
  total_wavenumber_slice = slice(0, save_coords.horizontal.modal_shape[1])
  if (
      coords.horizontal.total_wavenumbers
      < save_coords.horizontal.total_wavenumbers
  ) or (
      coords.horizontal.longitude_wavenumbers
      < save_coords.horizontal.longitude_wavenumbers
  ):
    raise ValueError('save_coords.horizontal larger than coords.horizontal')

  def downsample_fn(state: typing.PyTreeState) -> typing.PyTreeState:
    slice_fn = lambda x: x[..., lon_wavenumber_slice, total_wavenumber_slice]
    return pytree_utils.tree_map_over_nonscalars(slice_fn, state)

  return downsample_fn


def get_spectral_upsample_fn(
    coords: CoordinateSystem,
    save_coords: CoordinateSystem,
    expect_same_vertical: bool = True,
) -> Callable[[typing.PyTreeState], typing.PyTreeState]:
  """Returns function that upsamples state from `coords` to `save_coords`."""
  if expect_same_vertical and (coords.vertical != save_coords.vertical):
    raise ValueError('upsampling vertical resolution is not supported.')
  save_shape = save_coords.horizontal.modal_shape
  coords_shape = coords.horizontal.modal_shape
  lon_wavenumber_pad = (0, save_shape[0] - coords_shape[0])
  total_wavenumber_pad = (0, save_shape[1] - coords_shape[1])
  if (min(lon_wavenumber_pad) != 0) or (min(total_wavenumber_pad) != 0):
    raise ValueError('save_coords.horizontal smaller than coords.horizontal')
  tail_pad = (lon_wavenumber_pad, total_wavenumber_pad)

  def upsample_fn(state: typing.PyTreeState) -> typing.PyTreeState:
    pad_fn = lambda x: jnp.pad(x, ((0, 0),) * (x.ndim - 2) + tail_pad)
    return pytree_utils.tree_map_over_nonscalars(pad_fn, state)

  return upsample_fn


def get_spectral_interpolate_fn(
    source_coords: CoordinateSystem,
    target_coords: CoordinateSystem,
    expect_same_vertical: bool = True,
) -> Callable[[typing.PyTreeState], typing.PyTreeState]:
  """Returns modal interpolation_fn from `source_coords` to `target_coords`."""
  if (
      source_coords.horizontal.total_wavenumbers
      < target_coords.horizontal.total_wavenumbers
  ) and (
      source_coords.horizontal.longitude_wavenumbers
      < target_coords.horizontal.longitude_wavenumbers
  ):
    return get_spectral_upsample_fn(
        source_coords, target_coords, expect_same_vertical
    )
  elif (
      source_coords.horizontal.total_wavenumbers
      >= target_coords.horizontal.total_wavenumbers
  ) and (
      source_coords.horizontal.longitude_wavenumbers
      >= target_coords.horizontal.longitude_wavenumbers
  ):
    return get_spectral_downsample_fn(
        source_coords, target_coords, expect_same_vertical
    )
  else:
    raise ValueError(
        'Incompatible horizontal coordinates with shapes '
        f'{source_coords.horizontal.modal_shape}, '
        f'{target_coords.horizontal.modal_shape}'
    )


def get_nodal_shapes(
    inputs: typing.Pytree,
    coords: CoordinateSystem,
) -> typing.Pytree:
  """Returns a pytree with nodal shapes of arrays in `inputs`."""
  nodal_shape = coords.horizontal.nodal_shape
  array_shape_fn = lambda x: np.asarray(x.shape[:-2] + nodal_shape)
  scalar_shape_fn = lambda x: np.array([], dtype=int)
  return pytree_utils.tree_map_over_nonscalars(
      array_shape_fn, inputs, scalar_fn=scalar_shape_fn
  )


def get_modal_shapes(
    inputs: typing.Pytree,
    coords: CoordinateSystem,
) -> typing.Pytree:
  """Returns a pytree with modal shapes of arrays in `inputs`."""
  modal_shape = coords.horizontal.modal_shape
  array_shape_fn = lambda x: np.asarray(x.shape[:-2] + modal_shape)
  scalar_shape_fn = lambda x: np.array([], dtype=int)
  return pytree_utils.tree_map_over_nonscalars(
      array_shape_fn, inputs, scalar_fn=scalar_shape_fn
  )


# TODO(dkochkov) Consider moving these functions to spherical_harmonic module.


def maybe_to_nodal(
    fields: typing.Pytree,
    coords: CoordinateSystem,
) -> typing.Pytree:
  """Converts non-scalar elements in `fields` that are not nodal to nodal."""
  nodal_shapes = get_nodal_shapes(fields, coords)

  def to_nodal_fn(x):
    return coords.with_dycore_sharding(
        coords.horizontal.to_nodal(coords.with_dycore_sharding(x))
    )

  fn = lambda x, nodal: x if x.shape == tuple(nodal) else to_nodal_fn(x)
  return jax.tree_util.tree_map(fn, fields, nodal_shapes)


def maybe_to_modal(
    fields: typing.Pytree,
    coords: CoordinateSystem,
) -> typing.Pytree:
  """Converts non-scalar elements in `fields` that are not modal to modal."""
  modal_shapes = get_modal_shapes(fields, coords)

  def to_modal_fn(x):
    return coords.with_dycore_sharding(
        coords.horizontal.to_modal(coords.with_dycore_sharding(x))
    )

  fn = lambda x, modal: x if x.shape == tuple(modal) else to_modal_fn(x)
  return jax.tree_util.tree_map(fn, fields, modal_shapes)


def _map_over_matching_keys(
    inputs: dict[str, Any],
    fn: Callable[[typing.Array], typing.Array],
    keys_to_map_over: Sequence[str],
) -> dict[str, Any]:
  """Applies `fn` to values in `inputs` or sub-dictionaries for matching keys.

  Args:
    inputs: potentially nested dictionary to map over.
    fn: function to apply to elements in the `inputs` that have matching keys.
    keys_to_map_over: keys to which `fn` should be applied to.

  Returns:
    `inputs` with all elements that have keys matching an entry in
    `keys_to_map_over` transformed by function `fn`.
  """
  outputs = {}
  for k, v in inputs.items():
    if not isinstance(v, dict):
      outputs[k] = fn(v) if k in keys_to_map_over else v
    else:
      outputs[k] = _map_over_matching_keys(v, fn, keys_to_map_over)
  return outputs


def scale_levels_for_matching_keys(
    inputs: typing.Pytree,
    scales: typing.Array,
    keys_to_scale: Sequence[str] = tuple(),
) -> typing.Pytree:
  """Transforms `inputs` by scaling levels for keys that are in `keys_to_scale.

  Args:
    inputs: pytree of values that will be selectively scaled along levels.
    scales: scaling weights that will be applied.
    keys_to_scale: keys for which scaling operation is applied.

  Returns:
    pytree of the same structure of `inputs` with keys matching `keys_to_scale`
    scaled by `scales` along the level axis.
  """
  if scales.ndim != 1:
    raise ValueError(
        'scales must be 1d array of weights per level, got '
        f'array with shape {scales.shape}'
    )
  scales = scales[:, np.newaxis, np.newaxis]  # broadcasting shape.
  inputs, from_dict_fn = pytree_utils.as_dict(inputs)
  scale_fn = lambda x: x * scales
  inputs = _map_over_matching_keys(inputs, scale_fn, keys_to_scale)
  return from_dict_fn(inputs)



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for spherical_harmonic."""

import functools
from absl.testing import absltest
from absl.testing import parameterized
from dinosaur import spherical_harmonic
import jax
from jax import config
import jax.numpy as jnp
import numpy as np

config.update('jax_enable_x64', True)


def _function_0(lat, lon):
  return jnp.cos(lat) ** 4 * jnp.sin(3 * lon)


def _function_1(lat, lon):
  return jnp.cos(lat) ** 4 * jnp.sin(5 * lon) * jnp.cos(5 * lat)


def random_modal_state(grid, seed=0):
  _, l = grid.modal_mesh
  rs = np.random.RandomState(seed)
  array = rs.normal(size=grid.mask.shape)
  if np.issubdtype(grid.spherical_harmonics.modal_dtype, complex):
    array = array + 1j * rs.normal(size=grid.mask.shape)
  array *= grid.mask / (l + 1) ** 2
  return array


class SphericalHarmonicTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(
          longitude_nodes=64,
          latitude_nodes=32,
          longitude_wavenumbers=32,
          total_wavenumbers=32,
          latitude_spacing='gauss',
      ),
      dict(
          longitude_nodes=117,
          latitude_nodes=13,
          longitude_wavenumbers=45,
          total_wavenumbers=123,
          latitude_spacing='equiangular',
      ),
      dict(
          longitude_nodes=117,
          latitude_nodes=13,
          longitude_wavenumbers=45,
          total_wavenumbers=123,
          latitude_spacing='equiangular_with_poles',
      ),
  )
  def testBasisShapes(self, **params):
    """Tests that the arrays provided by `basis` have the expected shape."""
    spherical_harmonics = spherical_harmonic.RealSphericalHarmonics(**params)
    basis = spherical_harmonics.basis
    longitude_nodes = params['longitude_nodes']
    latitude_nodes = params['latitude_nodes']
    lon_waves, tot_waves = spherical_harmonics.modal_shape
    self.assertEqual((longitude_nodes, lon_waves), basis.f.shape)
    self.assertEqual(
        (lon_waves, latitude_nodes, tot_waves),
        basis.p.shape,
    )


class GridTest(parameterized.TestCase):

  @parameterized.product(
      wavenumbers=(32, 137),
      latitude_spacing=('gauss', 'equiangular'),
      impl=[
          spherical_harmonic.RealSphericalHarmonics,
          spherical_harmonic.FastSphericalHarmonics,
      ],
  )
  def testGridShape(self, wavenumbers, latitude_spacing, impl):
    """Check the nodal and modal shape attributes."""
    grid = spherical_harmonic.Grid.with_wavenumbers(
        wavenumbers,
        latitude_spacing=latitude_spacing,
        spherical_harmonics_impl=impl,
    )
    with self.subTest('nodal shape'):
      self.assertTupleEqual(grid.nodal_shape, grid.nodal_mesh[0].shape)
      self.assertTupleEqual(grid.nodal_shape, grid.nodal_mesh[1].shape)
      self.assertTupleEqual(
          grid.nodal_shape, (len(grid.nodal_axes[0]), len(grid.nodal_axes[1]))
      )

    with self.subTest('modal shape'):
      self.assertTupleEqual(grid.modal_shape, grid.modal_mesh[0].shape)
      self.assertTupleEqual(grid.modal_shape, grid.modal_mesh[1].shape)
      self.assertTupleEqual(
          grid.modal_shape, (len(grid.modal_axes[0]), len(grid.modal_axes[1]))
      )

  def testModalAxes(self):
    grid = spherical_harmonic.Grid(
        longitude_wavenumbers=4,
        total_wavenumbers=4,
        longitude_nodes=8,
        latitude_nodes=4,
    )
    m_expected = np.array([0, 1, -1, 2, -2, 3, -3])
    l_expected = np.array([0, 1, 2, 3])
    m_actual, l_actual = grid.modal_axes
    np.testing.assert_array_equal(m_expected, m_actual)
    np.testing.assert_array_equal(l_expected, l_actual)

  @parameterized.parameters(
      dict(longitude_offset=0.0),
      dict(longitude_offset=np.pi / 180),
      dict(longitude_offset=-np.pi / 180),
  )
  def test_longitudes(self, longitude_offset):
    grid = spherical_harmonic.Grid(
        longitude_wavenumbers=4,
        total_wavenumbers=4,
        longitude_nodes=8,
        latitude_nodes=4,
        longitude_offset=longitude_offset,
    )
    longitudes_expected = np.linspace(0, 2 * np.pi, 8, endpoint=False)
    longitudes_expected += grid.longitude_offset
    longitudes_actual, _ = grid.nodal_axes
    np.testing.assert_array_equal(longitudes_expected, longitudes_actual)

  def testConstructors(self):
    grid = spherical_harmonic.Grid.T21()
    self.assertEqual(grid.nodal_shape, (64, 32))
    self.assertEqual(grid.modal_shape, (43, 23))

    grid = spherical_harmonic.Grid.TL31()
    self.assertEqual(grid.nodal_shape, (64, 32))
    self.assertEqual(grid.modal_shape, (63, 33))

  @parameterized.parameters(
      dict(
          longitude_wavenumbers=32,
          total_wavenumbers=32,
          latitude_spacing='gauss',
          jit=False,
          seed=0,
          spherical_harmonics_impl=spherical_harmonic.RealSphericalHarmonics,
      ),
      dict(
          longitude_wavenumbers=64,
          total_wavenumbers=64,
          latitude_spacing='equiangular',
          jit=True,
          seed=0,
          spherical_harmonics_impl=spherical_harmonic.FastSphericalHarmonics,
      ),
      dict(
          longitude_wavenumbers=64,
          total_wavenumbers=64,
          latitude_spacing='equiangular',
          jit=True,
          seed=0,
          spherical_harmonics_impl=functools.partial(
              spherical_harmonic.FastSphericalHarmonics,
              base_shape_multiple=8,
              reverse_einsum_arg_order=True,
          ),
      ),
  )
  def testRoundTrip(
      self,
      longitude_wavenumbers,
      total_wavenumbers,
      latitude_spacing,
      jit,
      seed,
      spherical_harmonics_impl,
  ):
    """Tests that the modal -> nodal -> modal round trip is the identity."""
    longitude_nodes = 4 * longitude_wavenumbers
    latitude_nodes = 2 * total_wavenumbers

    grid = spherical_harmonic.Grid(
        longitude_nodes=longitude_nodes,
        latitude_nodes=latitude_nodes,
        longitude_wavenumbers=longitude_wavenumbers,
        total_wavenumbers=total_wavenumbers,
        latitude_spacing=latitude_spacing,
        spherical_harmonics_impl=spherical_harmonics_impl,
    )
    modal = random_modal_state(grid, seed=seed)
    modal[0, 0] = 0

    inverse_transform = grid.to_nodal
    transform = grid.to_modal
    if jit:
      inverse_transform = jax.jit(inverse_transform)
      transform = jax.jit(transform)

    nodal = inverse_transform(modal)
    reconstructed_modal = transform(nodal)
    np.testing.assert_allclose(modal, reconstructed_modal, atol=1e-5)

  @parameterized.product(
      wavenumbers=(32, 137, 255),
      latitude_spacing=('gauss', 'equiangular'),
      seed=(0,),
      impl=[
          spherical_harmonic.RealSphericalHarmonics,
          spherical_harmonic.FastSphericalHarmonics,
      ],
  )
  def testLaplacianRoundTrip(self, wavenumbers, latitude_spacing, seed, impl):
    """Similar that `inverse_laplacian` is the inverse of `laplacian`."""
    grid = spherical_harmonic.Grid.with_wavenumbers(
        wavenumbers,
        latitude_spacing=latitude_spacing,
        spherical_harmonics_impl=impl,
    )
    x = random_modal_state(grid, seed)
    x[0, 0] = 0
    y = grid.inverse_laplacian(grid.laplacian(x))
    np.testing.assert_allclose(x, y)

  @parameterized.product(
      (
          dict(latitude_spacing='gauss', atol=1e-5),
          dict(latitude_spacing='equiangular', atol=1e-5),
      ),
      wavenumbers=(64, 128),
      test_function=(_function_0, _function_1),
      impl=[
          spherical_harmonic.RealSphericalHarmonics,
          spherical_harmonic.FastSphericalHarmonics,
      ],
  )
  def testDerivatives(
      self, latitude_spacing, atol, wavenumbers, test_function, impl
  ):
    """Tests that `Grid` accurately computes derivatives."""
    grid = spherical_harmonic.Grid.with_wavenumbers(
        wavenumbers,
        latitude_spacing=latitude_spacing,
        spherical_harmonics_impl=impl,
    )
    lon, sin_lat = grid.nodal_mesh
    lat = np.arcsin(sin_lat)
    fx = test_function(lat, lon)

    with self.subTest(msg='sec_lat_d_dlat_cos2'):

      def cos2latf(lat, lon):
        return jnp.cos(lat) ** 2 * test_function(lat, lon)

      dcos2latf_dlat = jax.vmap(jax.vmap(jax.grad(cos2latf)))
      expected = dcos2latf_dlat(lat, lon) / grid.cos_lat
      actual = grid.to_nodal(grid.sec_lat_d_dlat_cos2(grid.to_modal(fx)))
      np.testing.assert_allclose(expected, actual, atol=atol)

    with self.subTest(msg='cos_lat_d_dlat'):
      df_dlat = jax.vmap(jax.vmap(jax.grad(test_function)))
      expected = df_dlat(lat, lon) * grid.cos_lat
      actual = grid.to_nodal(grid.cos_lat_d_dlat(grid.to_modal(fx)))
      np.testing.assert_allclose(expected, actual, atol=atol)

    with self.subTest(msg='d_dlon'):
      df_dlon = jax.vmap(jax.vmap(jax.grad(test_function, argnums=1)))
      expected = df_dlon(lat, lon)
      actual = grid.to_nodal(grid.d_dlon(grid.to_modal(fx)))
      np.testing.assert_allclose(expected, actual, atol=atol)

  @parameterized.parameters(
      dict(
          wavenumbers=85,
          latitude_spacing='equiangular',
          acceptable_norm_diff=10,
      ),
      dict(
          wavenumbers=85,
          latitude_spacing='gauss',
          acceptable_norm_diff=10,
      ),
      dict(
          wavenumbers=42,
          latitude_spacing='gauss',
          acceptable_norm_diff=5,
      ),
  )
  def testDerivativeArtifacts(
      self,
      wavenumbers,
      latitude_spacing,
      acceptable_norm_diff,
  ):
    """Tests that `Grid` computes derivatives without strong artifacts."""

    def test_function(lat, lon):
      """A hand-picked function that exposes derivative artifacts."""
      xs = jnp.linspace(0, 1.9, wavenumbers)
      ys = jnp.exp(jnp.sin(5 * xs) * 1.1 * xs**2 - 0.8 * xs**4)
      return sum(
          y * jnp.cos(lat * 4) ** 2 * jnp.cos(lon * (n % 4))
          for n, y in zip(np.arange(wavenumbers), (ys))
      )

    grid = spherical_harmonic.Grid.with_wavenumbers(
        wavenumbers, latitude_spacing=latitude_spacing
    )
    lon, sin_lat = grid.nodal_mesh
    lat = np.arcsin(sin_lat)
    fx = test_function(lat, lon)
    df_dlat = jax.vmap(jax.vmap(jax.grad(test_function)))
    expected = df_dlat(lat, lon) * grid.cos_lat

    with self.subTest(msg='artifact_in_cos_lat_d_dlat'):
      actual = grid.to_nodal(grid.cos_lat_d_dlat(grid.to_modal(fx)))
      error_norm = np.linalg.norm(actual - expected)
      self.assertGreater(error_norm, 10 * acceptable_norm_diff)

    with self.subTest(msg='no_artifact_in_cos_lat_grad'):
      _, actual = grid.to_nodal(grid.cos_lat_grad(grid.to_modal(fx)))
      error_norm = np.linalg.norm(actual - expected)
      self.assertLess(error_norm, acceptable_norm_diff)

  @parameterized.parameters(
      dict(grid=spherical_harmonic.Grid.with_wavenumbers(128), seed=0),
      dict(
          grid=spherical_harmonic.Grid(
              longitude_wavenumbers=64,
              total_wavenumbers=64,
              longitude_nodes=192,
              latitude_nodes=128,
              radius=2.6,
              latitude_spacing='equiangular',
          ),
          seed=0,
      ),
      dict(
          grid=spherical_harmonic.Grid.with_wavenumbers(128, radius=0.3), seed=0
      ),
  )
  def testLaplacian(self, grid, seed):
    """Tests that computing the Laplacian in 2 ways gives identical results."""
    # The test runs on random input with high wavenumber components damped.
    x = random_modal_state(grid, seed)
    x[0, 0] = 0
    # Taking the derivative twice will give inaccurate results for the highest
    # total wavenumber, so we trim these values for testing.
    x[:, -1] = 0

    # Compute Laplacian using eigenvalues.
    laplacian0 = grid.laplacian(x)

    # Compute Laplacian using
    # Δx =  ∇ · [cosθ ((cosθ ∇x) / cos²θ)]
    # where θ is latitude.
    # `x` has no top wavenumbers, so it's safe to skip clipping once.
    cos_lat_grad = grid.cos_lat_grad(x, clip=False)
    sec_lat_grad = grid.to_modal(grid.to_nodal(cos_lat_grad) * grid.sec2_lat)
    laplacian1 = grid.div_cos_lat(sec_lat_grad)

    np.testing.assert_allclose(laplacian0, laplacian1, atol=1e-10)

  @parameterized.parameters(
      dict(
          grid=spherical_harmonic.Grid.with_wavenumbers(128), atol=1e-10, seed=0
      ),
      dict(
          grid=spherical_harmonic.Grid.with_wavenumbers(
              128,
              spherical_harmonics_impl=spherical_harmonic.FastSphericalHarmonics,
          ),
          atol=1e-11,
          seed=0,
      ),
      dict(
          grid=spherical_harmonic.Grid.with_wavenumbers(
              128,
              spherical_harmonics_impl=functools.partial(
                  spherical_harmonic.FastSphericalHarmonics,
                  transform_precision='float32',
              ),
          ),
          atol=1e-11,
          seed=0,
      ),
      dict(
          grid=spherical_harmonic.Grid(
              longitude_wavenumbers=64,
              total_wavenumbers=64,
              longitude_nodes=192,
              latitude_nodes=128,
              radius=3.2,
              latitude_spacing='equiangular',
          ),
          atol=1e-5,
          seed=0,
      ),
      dict(
          grid=spherical_harmonic.Grid.with_wavenumbers(128, radius=0.54),
          atol=1e-10,
          seed=0,
      ),
  )
  def testVorticityStreamVelocityRoundTrip(self, grid, atol, seed):
    """Tests the vorticity -> stream -> velocity round trip is the identity."""
    # Choose a random vorticity 𝜻.
    vorticity = random_modal_state(grid, seed)
    vorticity[0, 0] = 0
    vorticity[:, -1] = 0

    # Solve for the stream function ∇²ѱ = 𝜻
    stream = grid.inverse_laplacian(vorticity)

    # Compute the velocity v = k x 𝝯ѱ
    # `stream` has no top wavenumbers, so it's safe to skip clipping once.
    cos_lat_v = jnp.stack(grid.k_cross(grid.cos_lat_grad(stream, clip=False)))
    sec_lat_v = grid.to_modal(grid.sec2_lat * grid.to_nodal(cos_lat_v))
    div_v = grid.div_cos_lat(sec_lat_v)

    # We expect the velocity to be divergence-free.
    np.testing.assert_allclose(div_v, 0, atol=atol)

    # Reconstruct the vorticity 𝜻 = ∇ x v.
    reconstructed_vorticity = grid.curl_cos_lat(sec_lat_v)

    np.testing.assert_allclose(vorticity, reconstructed_vorticity, atol=atol)

  @parameterized.parameters(
      dict(
          grid=spherical_harmonic.Grid.with_wavenumbers(128), atol=1e-10, seed=0
      ),
      dict(
          grid=spherical_harmonic.Grid(
              longitude_wavenumbers=64,
              total_wavenumbers=64,
              longitude_nodes=192,
              latitude_nodes=128,
              radius=3.2,
              latitude_spacing='equiangular',
          ),
          atol=1e-5,
          seed=0,
      ),
      dict(
          grid=spherical_harmonic.Grid.with_wavenumbers(128, radius=0.54),
          atol=1e-10,
          seed=0,
      ),
  )
  def testDivergencePotentialVelocityRoundTrip(self, grid, atol, seed):
    """Tests the div -> potential -> velocity round trip is the identity."""
    # Choose a random divergence D.
    divergence = random_modal_state(grid, seed)
    divergence[0, 0] = 0
    divergence[:, -1] = 0

    # Solve for the velocity potential ∇²ɸ = D.
    potential = grid.inverse_laplacian(divergence)

    # Compute the velocity v = 𝝯ɸ
    # `potential` has no top wavenumbers, so it's safe to skip clipping once.
    cos_lat_v = grid.cos_lat_grad(potential, clip=False)
    sec_lat_v = grid.to_modal(grid.sec2_lat * grid.to_nodal(cos_lat_v))
    curl_v = grid.curl_cos_lat(sec_lat_v)

    # We expect the velocity to be curl-free.
    np.testing.assert_allclose(curl_v, 0, atol=atol)

    # Reconstruct the divergence D = ∇ · v.
    reconstructed_divergence = grid.div_cos_lat(sec_lat_v).at[:, -1].set(0)

    np.testing.assert_allclose(divergence, reconstructed_divergence, atol=atol)

  @parameterized.parameters(
      dict(wavenumbers=64, latitude_spacing='gauss', radius=1.3),
      dict(wavenumbers=128, latitude_spacing='equiangular', radius=2.5),
      dict(wavenumbers=256, latitude_spacing='gauss', radius=1.0),
  )
  def testIntegrationSurfaceArea(self, wavenumbers, latitude_spacing, radius):
    grid = spherical_harmonic.Grid.with_wavenumbers(
        wavenumbers, latitude_spacing=latitude_spacing, radius=radius
    )
    ones = jnp.ones([grid.longitude_nodes, grid.latitude_nodes])
    quadrature_surface_area = grid.integrate(ones)
    expected_surface_area = 4 * np.pi * radius**2  # A = 4πr²
    np.testing.assert_array_almost_equal(
        quadrature_surface_area, expected_surface_area
    )

  @parameterized.product(
      params=[
          dict(wavenumbers=64, l=0, m=0),
          dict(wavenumbers=128, l=32, m=17),
          dict(wavenumbers=256, l=174, m=95),
      ],
      impl=[
          spherical_harmonic.RealSphericalHarmonics,
          spherical_harmonic.FastSphericalHarmonics,
      ],
  )
  def testIntegrationSphericalHarmonics(self, params, impl):
    wavenumbers = params['wavenumbers']
    l = params['l']
    m = params['m']
    grid = spherical_harmonic.Grid.with_wavenumbers(
        wavenumbers, spherical_harmonics_impl=impl
    )
    x = jnp.zeros_like(grid.mask).at[m, l].set(1)
    z = grid.to_nodal(x)
    integral = grid.integrate(z**2)
    np.testing.assert_allclose(integral, 1, atol=1e-10)

  def testClipWavenumber(self):
    state = {'u': np.ones((4, 3)), 'time': 1.0}
    grid = spherical_harmonic.Grid(longitude_wavenumbers=2, total_wavenumbers=3)
    clipped = grid.clip_wavenumbers(state)
    expected_u = np.array([[1, 1, 0], [1, 1, 0], [1, 1, 0], [1, 1, 0]])
    np.testing.assert_array_equal(clipped['u'], expected_u)
    self.assertEqual(clipped['time'], 1.0)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A vertical coordinate system for a system with layers of fixed density."""

from __future__ import annotations

import dataclasses
import numpy as np


@dataclasses.dataclass(frozen=True)
class LayerCoordinates:
  """Vertical coordinate system for a multi-layer shallow water equations.

  A description of a discrete, constant density vertical layer. Layers are
  indexed from the "top"  of the atmosphere to the surface of the earth.

  Attributes:
    layers: the number of layers.
    centers: indices of layers.
  """
  layers: int

  @property
  def centers(self) -> np.ndarray:
    return np.arange(self.layers)

  def asdict(self):
    return dataclasses.asdict(self)




# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions that operate on pytrees."""
from collections import abc
import dataclasses
import functools
from typing import Any, Callable, Sequence

from dinosaur import typing

import jax
import jax.numpy as jnp
import numpy as np


tree_map = jax.tree_util.tree_map


def pack_pytree(pytree: typing.Pytree, axis: int = -3) -> typing.Array:
  """Packs `pytree` by concatenating leaves along `axis`."""
  flat, _ = jax.tree_util.tree_flatten(pytree)
  if not flat:
    return None  # pytype: disable=bad-return-type  # jax-ndarray
  packed = jnp.concatenate(flat, axis)
  return packed


def unpack_to_pytree(
    array: typing.Array,
    pytree_of_shapes: typing.Pytree,
    axis: int = -3
) -> typing.Pytree:
  """Unpacks an `array` into a pytree with shapes `pytree_of_shapes`."""
  shapes, tree_def = jax.tree_util.tree_flatten(pytree_of_shapes)
  splits = np.cumsum(np.array([x[axis] for x in shapes]))[:-1]
  split = jnp.split(array, splits, axis)
  return jax.tree_util.tree_unflatten(tree_def, split)


def stack_pytree(pytree: typing.Pytree, axis: int = 0) -> typing.Array:
  """Stacks `pytree` by stacking leaves along a new axis."""
  flat, _ = jax.tree_util.tree_flatten(pytree)
  if not flat:
    return None  # pytype: disable=bad-return-type  # jax-ndarray
  stacked = jnp.stack(flat, axis)
  return stacked


def unstack_to_pytree(
    array: typing.Array,
    pytree_of_shapes: typing.Pytree,
    axis: int = 0
) -> typing.Pytree:
  """Unstacks an `array` into a pytree with shapes `pytree_of_shapes`."""
  _, tree_def = jax.tree_util.tree_flatten(pytree_of_shapes)
  # `array` is split along `axis` and resulting singleton dimension is removed
  split = jnp.split(array, array.shape[axis], axis)
  split = tree_map(lambda x: jnp.squeeze(x, axis=axis), split)
  return jax.tree_util.tree_unflatten(tree_def, split)


def tree_map_where(
    condition_fn: Callable[[typing.Array], typing.Array],
    f: Callable[[typing.Array], typing.Array],
    g: Callable[[typing.Array], typing.Array],
    x: typing.Pytree,
) -> typing.Pytree:
  """Map `f` over pytree leaves if condition_fn(leaf), else use `g`."""
  def tm_where(x: typing.Array) -> typing.Array:
    return f(x) if condition_fn(x) else g(x)
  return tree_map(tm_where, x)


def tree_map_over_nonscalars(
    f: Callable[[typing.Array], typing.Array],
    x: typing.Pytree,
    *,
    scalar_fn: Callable[[typing.Array], typing.Array] = lambda x: x,
    backend: str = 'jax',
) -> typing.Pytree:
  """Map `f` over nonscalar pytree leaves, but use `scalar_fn` on scalars."""
  as_array_fn = {'jax': jnp.asarray, 'numpy': np.asarray}[backend]
  def g(x: typing.Array) -> typing.Array:
    x = as_array_fn(x)
    return f(x) if x.ndim else scalar_fn(x)
  return tree_map(g, x)


def shape_structure(inputs):
  """Returns `inputs` with leafs replaced by arrays of corresponding shapes."""
  return tree_map(lambda x: np.asarray(x.shape), inputs)


def _normalize_axis(axis: int, ndim: int) -> int:
  """Validates and returns positive `axis` value."""
  if not -ndim <= axis < ndim:
    raise ValueError(f'invalid axis {axis} for ndim {ndim}')
  if axis < 0:
    axis += ndim
  return axis


def slice_along_axis(
    inputs: typing.Pytree,
    axis: int,
    idx: int | slice,
    expect_same_dims: bool = False,
) -> typing.Pytree:
  """Returns slice of `inputs` defined by `idx` along axis `axis`.

  Args:
    inputs: pytree to slice.
    axis: axis along which to slice `inputs`.
    idx: index or slice along axis `axis` that is returned.
    expect_same_dims: whether all arrays should have same number of dimensions.

  Returns:
    Slice of `inputs` defined by `idx` along axis `axis`. If idx is an int,
    the axis is dropped. If idx is a slice, the axis is handled accordingly.
  """
  arrays, tree_def = jax.tree_util.tree_flatten(inputs)
  ndims = set(a.ndim for a in arrays)
  if expect_same_dims and len(ndims) != 1:
    raise ValueError('arrays in `inputs` expected to have same ndims, but have '
                     f'{ndims}. To allow this, pass expect_same_dims=False')
  elif axis < 0:
    raise ValueError(f'Using {axis=} is error-prone if expect_same_dims=False')
  sliced = []
  for array in arrays:
    ndim = array.ndim
    slc = tuple(idx if j == _normalize_axis(axis, ndim) else slice(None)
                for j in range(ndim))
    sliced.append(array[slc])
  return jax.tree_util.tree_unflatten(tree_def, sliced)


def split_along_axis(
    inputs: typing.Pytree,
    split_idx: int,
    axis: int,
    expect_same_dims: bool = False,
) -> tuple[typing.Pytree, typing.Pytree]:
  """Returns 2 pytrees formed by splitting `inputs` along `axis` at `split_idx`.

  Args:
    inputs: pytree to split.
    split_idx: index along `axis` where the second split starts.
    axis: axis along which to split the `inputs`.
    expect_same_dims: whether all arrays should have same number of dimensions.

  Returns:
    Tuple of slices of `inputs` split along `axis` at `split_idx`.
  """

  first_slice = slice_along_axis(
      inputs, axis, slice(0, split_idx), expect_same_dims)
  second_slice = slice_along_axis(
      inputs, axis, slice(split_idx, None), expect_same_dims)
  return first_slice, second_slice


def split_axis(
    inputs: typing.Pytree,
    axis: int,
    keep_dims: bool = False,
) -> tuple[typing.Pytree, ...]:
  """Splits `inputs` along `axis`.

  Args:
    inputs: pytree to be split.
    axis: axis along which to split the `inputs`.
    keep_dims: whether to keep `axis` dimension.

  Returns:
    Tuple of pytrees that correspond to slices of `inputs` along `axis`. The
    `axis` dimension is removed if `squeeze is set to True.

  Raises:
    ValueError: if arrays in `inputs` don't have unique size along `axis`.
  """
  arrays, tree_def = jax.tree_util.tree_flatten(inputs)
  axis_shapes = set(a.shape[axis] for a in arrays)
  if len(axis_shapes) != 1:
    raise ValueError(f'Arrays must have equal sized axis but got {axis_shapes}')
  axis_shape, = axis_shapes
  splits = [jnp.split(a, axis_shape, axis=axis) for a in arrays]
  if not keep_dims:
    splits = tree_map(lambda a: jnp.squeeze(a, axis), splits)
  splits = zip(*splits)
  return tuple(jax.tree_util.tree_unflatten(tree_def, leaves)
               for leaves in splits)


def concat_along_axis(
    pytrees: Sequence[typing.Pytree],
    axis: int
) -> typing.Pytree:
  """Concatenates `pytrees` along `axis`."""
  concat_leaves_fn = lambda *args: jnp.concatenate(args, axis)
  return tree_map(concat_leaves_fn, *pytrees)


def as_dict(inputs: typing.Pytree) -> typing.Pytree:
  """Returns a dict representation of `inputs` and a from_dict_fn."""
  return_type = type(inputs)
  if dataclasses.is_dataclass(inputs):
    inputs = inputs.asdict()
  else:
    if return_type != dict:
      raise ValueError(f'Inputs of type {return_type} are not supported.')

  from_dict_fn = lambda dict_inputs: return_type(**dict_inputs)
  return inputs, from_dict_fn


def none_to_zeros(
    tree: typing.Pytree,
    reference_tree: typing.Pytree,
) -> typing.Pytree:
  """Returns tree where `None` is replaced with zeros_like of reference_tree."""
  return tree_map(
      lambda x, y: jnp.zeros_like(y) if x is None else x, tree, reference_tree)


@dataclasses.dataclass(frozen=True)
class _HashableNDArrayWrapper:
  shape: tuple[int, ...]
  dtype: np.dtype
  data: bytes


def _hash_leaf(x: typing.Pytree) -> abc.Hashable:
  if isinstance(x, (jax.Array, np.ndarray)):
    return _HashableNDArrayWrapper(x.shape, x.dtype, x.tobytes())
  else:
    return x


def tree_hashable(x: typing.Pytree) -> abc.Hashable:
  """Convert a pytree into something hashable."""
  values, treedef = jax.tree_util.tree_flatten(x)
  values = tuple(map(_hash_leaf, values))
  return values, treedef


def tree_cache(func):
  """Like functools.cache, but hashes with tree_hashable.

  Example usage::

    import numpy as np

    @tree_cache
    def f(x):
      print('caching')
      return x

    >>> f({'a': 1, 'b': np.arange(3)})
    caching
    {'a': 1, 'b': array([0, 1, 2])}
    >>> f({'a': 1, 'b': np.arange(3)})
    {'a': 1, 'b': array([0, 1, 2])}

  Args:
    func: function to cache.

  Returns:
    Function where cached results are reused.
  """
  results = {}

  @functools.wraps(func)
  def wrapper(*args, **kwargs):
    key = tree_hashable((args, kwargs))
    if key in results:
      return results[key]
    result = results[key] = func(*args, **kwargs)
    return result

  return wrapper


def flatten_dict(
    input_dict: dict[str, Any],
    prefix: str = '',
    sep: str = '&',
) -> tuple[dict[str, Any], tuple[str, ...]]:
  """Flattens potentially nested `input_dict`."""
  items = []
  empty_keys = []
  for k, v in input_dict.items():
    if sep in k:
      raise ValueError(f'Key {k} contains {sep=}. Use different name or sep.')
    new_key = prefix + sep + k if prefix else k
    if isinstance(v, dict) and v:
      sub_dict, sub_empty_keys = flatten_dict(v, new_key, sep=sep)
      items.extend(sub_dict.items())
      empty_keys.extend(sub_empty_keys)
    elif isinstance(v, dict) and not v:
      empty_keys.append(new_key)
    else:
      items.append((new_key, v))
  unique_keys, counts = np.unique(
      np.array([x[0] for x in items]), return_counts=True)
  if (counts > 1).any():
    raise ValueError(f'got duplicate keys {unique_keys[counts > 1]}')
  unique_empty_keys, counts = np.unique(
      np.array([x[0] for x in empty_keys]), return_counts=True)
  if (counts > 1).any():
    raise ValueError(f'got duplicate keys {unique_empty_keys[counts > 1]}')
  return dict(items), tuple(empty_keys)


def unflatten_dict(
    flat_dict: dict[str, Any],
    empty_keys: tuple[str, ...] = tuple(),
    sep: str = '&',
) -> dict[str, Any]:
  """Unflattens `flat_dict` with structure specified with separataion `sep`."""
  result = dict()
  empty_key_dict = {k: {} for k in empty_keys}
  for key, value in (flat_dict | empty_key_dict).items():
    sub_keys = key.split(sep)
    sub_dict = result
    for sub_key in sub_keys[:-1]:
      if sub_key in sub_dict:
        sub_dict = sub_dict[sub_key]
      else:
        sub_dict[sub_key] = dict()
        sub_dict = sub_dict[sub_key]
    sub_dict[sub_keys[-1]] = value
  return result


def replace_with_matching_or_default(
    x: dict[str, Any],
    replace: dict[str, Any],
    default: Any = None,
    check_used_all_replace_keys: bool = True,
) -> dict[str, Any]:
  """Returns `x` structure with leaves from `replace` or `default`."""
  flat_x, empty_keys = flatten_dict(x)
  flat_replace, _ = flatten_dict(replace)
  if check_used_all_replace_keys:
    unused_replace_keys = set(flat_replace.keys()) - set(flat_x.keys())
    if unused_replace_keys:
      raise ValueError(f'Keys {unused_replace_keys} not present in {x.keys()=}')
  flat_result = {k: flat_replace.get(k, default) for k in flat_x.keys()}
  return unflatten_dict(flat_result, empty_keys)



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defined commonly used types in the codebase."""

import dataclasses
from typing import Any, Callable, Generic, Mapping, TypeVar

from dinosaur import scales

import jax.numpy as jnp
import numpy as np
import tree_math


Array = np.ndarray | jnp.ndarray
ArrayOrArrayTuple = Array | tuple[Array, ...]
Numeric = float | int | Array
Quantity = scales.Quantity
PRNGKeyArray = Any

# TODO(dkochkov): Rename PyTreeXXX to XXX once old types are pruned.
PyTreeState = TypeVar('PyTreeState')
Pytree = Any
PyTreeMemory = Pytree
PyTreeDiagnostics = Pytree

AuxFeatures = dict[str, Any]
DataState = dict[str, Any]
ForcingData = dict[str, Any]


@dataclasses.dataclass(eq=True, order=True, frozen=True)
class KeyWithCosLatFactor:
  """Class describing a key by `name` and an integer `factor_order`."""
  name: str
  factor_order: int
  filter_strength: float = 0.0


@tree_math.struct
class RandomnessState:
  """Representation of random states on the sphere.

  Attributes:
    core: internal representation of the random state.
    nodal_value: random field values in the nodal representation.
    modal_value: random field values in the modal representation.
    prng_key: underlying PRNG key.
    prng_step: optional iteration counter associated with PRNG key access, to
      allow for avoiding the pattern of iterative splitting the same key, which
      has poor statistical properties. The recommended pattern for generating a
      new PRNG key is `jax.random.fold_in(state.prng_key, state.prng_step)`.
  """
  core: Pytree | None = None
  nodal_value: Pytree | None = None
  modal_value: Pytree | None = None
  prng_key: PRNGKeyArray | None = None
  prng_step: int | None = None


@tree_math.struct
class ModelState(Generic[PyTreeState]):
  """PyTreeState decomposed into deterministic and perturbation components.

  A stochastic model advances by maintaining a model state and perturbations.

  Attributes:
    state: Prognostic variables describing the state of the atmosphere.
    memory: Optional model fields/predictions providing past time context.
    diagnostics: Optional diagnostic values computed in the model space.
    randomness: An optional random field that is used to stochastically perturb
      the advance step of the model.
  """
  state: PyTreeState
  memory: Pytree = dataclasses.field(default=None)
  diagnostics: Pytree = dataclasses.field(default_factory=dict)
  randomness: RandomnessState = dataclasses.field(
      default_factory=RandomnessState
  )


@tree_math.struct
class TrajectoryRepresentations:
  """Dataclass that holds trajectories in all default representations."""
  data_nodal_trajectory: Pytree
  data_modal_trajectory: Pytree
  model_nodal_trajectory: Pytree
  model_modal_trajectory: Pytree

  def get_representation(self, *, is_nodal: bool, is_encoded: bool) -> Pytree:
    """Retrieves representation based on `is_nodal`, `is_encoded` booleans."""
    binary_nodal_encoded_dict = {
        (True, True): self.model_nodal_trajectory,
        (True, False): self.data_nodal_trajectory,
        (False, True): self.model_modal_trajectory,
        (False, False): self.data_modal_trajectory,
    }
    return binary_nodal_encoded_dict[(is_nodal, is_encoded)]

# TODO(dkochkov) unify State and PyTreeState and integrators.
State = TypeVar('State')
StateFn = Callable[[State], State]
InverseFn = Callable[[State, jnp.ndarray], State]
StepFn = Callable[[State, State], State]
FilterFn = Callable[[State, State, State], tuple[State, State]]

ScanFn = Callable[..., Any]
PytreeFn = Callable[[Pytree], Pytree]
PyTreeTermsFn = Callable[[PyTreeState], PyTreeState]
PyTreeInverseFn = Callable[[PyTreeState, Numeric], PyTreeState]
TimeStepFn = Callable[[PyTreeState], PyTreeState]
PyTreeFilterFn = Callable[[PyTreeState], PyTreeState]
PyTreeStepFilterFn = Callable[[PyTreeState, PyTreeState], PyTreeState]
PyTreeStepFilterModule = Callable[..., PyTreeStepFilterFn]

Forcing = Pytree
ForcingFn = Callable[[ForcingData, float], Forcing]
ForcingModule = Callable[..., ForcingFn]

PostProcessFn = Callable[..., Any]
Params = Mapping[str, Mapping[str, Array]] | None

StepFn = Callable[[PyTreeState, Forcing | None], PyTreeState]
StepModule = Callable[..., StepFn]
CorrectorFn = Callable[
    [PyTreeState, PyTreeState | None, Forcing | None], PyTreeState
]
CorrectorModule = Callable[..., CorrectorFn]
ParameterizationFn = Callable[
    [
        PyTreeState,
        PyTreeMemory | None,
        PyTreeDiagnostics | None,
        RandomnessState | None,
        Forcing | None,
    ],
    PyTreeState,
]
ParameterizationModule = Callable[..., ParameterizationFn]
TrajectoryFn = Callable[..., tuple[Any, Any]]
TransformFn = Callable[[Pytree], Pytree]
TransformModule = Callable[..., TransformFn]

GatingFactory = Callable[..., Callable[[Array, Array], Array]]
TowerFactory = Callable[..., Callable[..., Any]]
LayerFactory = Callable[..., Callable[..., Any]]

EmbeddingFn = Callable[
    [
        Pytree,
        PyTreeMemory | None,
        PyTreeDiagnostics | None,
        RandomnessState | None,
        Forcing | None,
    ],
    Pytree,
]
EmbeddingModule = Callable[..., EmbeddingFn]



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for radiation."""

import datetime

from absl.testing import absltest
from absl.testing import parameterized

from dinosaur import coordinate_systems
from dinosaur import primitive_equations
from dinosaur import radiation
from dinosaur import scales
from dinosaur import sigma_coordinates
from dinosaur import spherical_harmonic

import jax
from jax import config
import jax.numpy as jnp
import numpy as np

units = scales.units
config.parse_flags_with_absl()

TWOPI = 2 * jnp.pi


class RadiationTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(when=np.datetime64('1980-01-01T00:00:00.000000000'),
           expected=datetime.datetime(1980, 1, 1, 0, 0)),
      dict(when=np.datetime64('2001-02-03T04:05'),
           expected=datetime.datetime(2001, 2, 3, 4, 5)),
      dict(when=np.datetime64('1982-05-04'),
           expected=datetime.datetime(1982, 5, 4, 0, 0)),
  )
  def test_datetime64_to_datetime(self, when, expected):
    actual = radiation.datetime64_to_datetime(when)
    self.assertEqual(actual, expected)

  @parameterized.parameters(
      dict(when=datetime.datetime(1984, 1, 1), expected_days=366),
      dict(when=datetime.datetime(1985, 1, 1), expected_days=365),
      dict(when=datetime.datetime(2020, 1, 1), expected_days=366),
      dict(when=datetime.datetime(2001, 1, 1), expected_days=365),
  )
  def test_days_in_year(self, when, expected_days):
    actual_days = radiation.days_in_year(when)
    self.assertEqual(actual_days, expected_days)

  @parameterized.parameters(
      # 1980 is a leap year
      dict(when=datetime.datetime(1980, 1, 1, 0, 0),
           exp_orbital_phase=0,
           exp_synodic_phase=0),
      dict(when=datetime.datetime(1980, 1, 1, 12, 0),
           exp_orbital_phase=0.5 * TWOPI / 366,
           exp_synodic_phase=jnp.pi),
      dict(when=datetime.datetime(1980, 1, 1, 23, 59),
           exp_orbital_phase=(1439 / 1440) * TWOPI / 366,
           exp_synodic_phase=1439 * TWOPI / 1440),
      dict(when=datetime.datetime(1980, 1, 1, 23, 59),
           exp_orbital_phase=(1439 / 1440) * TWOPI / 366,
           exp_synodic_phase=1439 * TWOPI / 1440),
      dict(when=datetime.datetime(1980, 5, 1, 0, 0),
           exp_orbital_phase=(31 + 29 + 31 + 30) * TWOPI / 366,
           exp_synodic_phase=0),
      dict(when=datetime.datetime(1980, 12, 31, 23, 59),
           exp_orbital_phase=TWOPI - (1 / 1440) * TWOPI / 366,
           exp_synodic_phase=1439 * TWOPI / 1440),
      # 1981 is not a leap year
      dict(when=datetime.datetime(1981, 1, 1, 0, 0),
           exp_orbital_phase=0,
           exp_synodic_phase=0),
      dict(when=datetime.datetime(1981, 1, 1, 12, 0),
           exp_orbital_phase=0.5 * TWOPI / 365,
           exp_synodic_phase=jnp.pi),
      dict(when=datetime.datetime(1981, 1, 1, 23, 59),
           exp_orbital_phase=(1439 / 1440) * TWOPI / 365,
           exp_synodic_phase=1439 * TWOPI / 1440),
      dict(when=datetime.datetime(1981, 1, 1, 23, 59),
           exp_orbital_phase=(1439 / 1440) * TWOPI / 365,
           exp_synodic_phase=1439 * TWOPI / 1440),
      dict(when=datetime.datetime(1981, 5, 1, 0, 0),
           exp_orbital_phase=(31 + 28 + 31 + 30) * TWOPI / 365,
           exp_synodic_phase=0),
      dict(when=datetime.datetime(1981, 12, 31, 23, 59),
           exp_orbital_phase=TWOPI - (1 / 1440) * TWOPI / 365,
           exp_synodic_phase=1439 * TWOPI / 1440),
      # The distant future
      dict(when=datetime.datetime(2022, 5, 1, 0, 0),
           exp_orbital_phase=(31 + 28 + 31 + 30) * TWOPI / 365,
           exp_synodic_phase=0),
      dict(when=datetime.datetime(2045, 12, 31, 23, 59),
           exp_orbital_phase=TWOPI - (1 / 1440) * TWOPI / 365,
           exp_synodic_phase=1439 * TWOPI / 1440),
  )
  def test_datetime_to_orbital_time(
      self, when, exp_orbital_phase, exp_synodic_phase):
    actual = radiation.datetime_to_orbital_time(when)
    self.assertAlmostEqual(actual.orbital_phase, exp_orbital_phase)
    self.assertAlmostEqual(actual.synodic_phase, exp_synodic_phase)

  def test_get_direct_solar_irradiance_no_units(self):
    flux = radiation.get_direct_solar_irradiance(
        orbital_phase=jnp.linspace(0, TWOPI, 4, endpoint=False),
        mean_irradiance=1.2,
        variation=0.3,
        perihelion=0)
    np.testing.assert_allclose(flux, [1.5, 1.2, 0.9, 1.2])

  def test_direct_solar_irradiance_pint_units(self):
    flux = radiation.get_direct_solar_irradiance(
        orbital_phase=jnp.linspace(0, TWOPI, 4, endpoint=False),
        mean_irradiance=1.2 * units.W / units.meter**2,
        variation=0.3 * units.W / units.meter**2,
        perihelion=0)
    np.testing.assert_allclose(
        flux, jnp.array([1.5, 1.2, 0.9, 1.2]) * units.W / units.meter**2)

  # Days of the year when equation of time is nearly zero
  @parameterized.parameters(
      (datetime.datetime(2000, 4, 16),),
      (datetime.datetime(2000, 6, 14),),
      (datetime.datetime(2000, 8, 31),),
      (datetime.datetime(2000, 12, 25),),
  )
  def test_equation_of_time(self, when):
    ot = radiation.datetime_to_orbital_time(when)
    delta_phase = radiation.equation_of_time(ot.orbital_phase)
    np.testing.assert_allclose(delta_phase, 0, atol=0.005)


def _get_expected_value_modulo_2pi(expected, actual):
  """Returns expected values for an approximate comparison modulo 2pi."""
    # Use np.fmod to avoid loss of precision with jnp.fmod.
  expected = np.fmod(expected, TWOPI)
  expected = jnp.where(expected < actual - jnp.pi, expected + TWOPI, expected)
  expected = jnp.where(expected > actual + jnp.pi, expected - TWOPI, expected)
  return expected


class SolarRadiationTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.coords = coordinate_systems.CoordinateSystem(
        horizontal=spherical_harmonic.Grid.T85(),
        vertical=sigma_coordinates.SigmaCoordinates.equidistant(layers=8))
    self.physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    self.reference_datetime = radiation.WB_REFERENCE_DATETIME

  def test_radiation_shapes(self):
    solar_radiation = radiation.SolarRadiation(
        self.coords, self.physics_specs, self.reference_datetime)
    with self.subTest('radiation_flux'):
      radiation_flux_fn = jax.jit(solar_radiation.radiation_flux)
      tisr = radiation_flux_fn(time=0)
      self.assertEqual(tisr.shape, self.coords.horizontal.nodal_shape)
    with self.subTest('solar_hour_angle'):
      solar_hour_angle_fn = jax.jit(solar_radiation.solar_hour_angle)
      solor_hour_angle = solar_hour_angle_fn(time=0)
      self.assertEqual(solor_hour_angle.shape,
                       self.coords.horizontal.nodal_shape)

  def test_radiation_flux_nondimensionalization(self):
    solar_radiation = radiation.SolarRadiation(
        self.coords, self.physics_specs, self.reference_datetime)
    perihelion_datetime = datetime.datetime(1979, 1, 3)
    time = solar_radiation.datetime_to_time(perihelion_datetime)
    radiation_flux_fn = jax.jit(solar_radiation.radiation_flux)
    actual_max = np.max(radiation_flux_fn(time))
    # At perihelion, irradiance is at its highest
    expected_max = self.physics_specs.nondimensionalize(
        radiation.TOTAL_SOLAR_IRRADIANCE + radiation.SOLAR_IRRADIANCE_VARIATION)
    np.testing.assert_allclose(actual_max, expected_max, rtol=5e-5)

  @parameterized.parameters(
      dict(reference_datetime=radiation.WB_REFERENCE_DATETIME,
           days_until_aphelion=3 + radiation.DAYS_PER_YEAR / 2),
      dict(reference_datetime=radiation.WB_REFERENCE_DATETIME,
           days_until_aphelion=3 - radiation.DAYS_PER_YEAR / 2),
      dict(reference_datetime=datetime.datetime(1979, 7, 4),
           days_until_aphelion=0),
      dict(reference_datetime=datetime.datetime(1980, 7, 3),
           days_until_aphelion=0),
      dict(reference_datetime=datetime.datetime(2022, 5, 4),
           days_until_aphelion=61),
  )
  def test_reference_datetime(self, reference_datetime, days_until_aphelion):
    # Check that combinations of reference_datetime and time work correctly
    solar_radiation = radiation.SolarRadiation(
        self.coords, self.physics_specs, reference_datetime)
    time = self.physics_specs.nondimensionalize(days_until_aphelion * units.day)
    radiation_flux_fn = jax.jit(solar_radiation.radiation_flux)
    actual_max = np.max(radiation_flux_fn(time))
    # At aphelion, irradiance is at its lowest
    expected_max = self.physics_specs.nondimensionalize(
        radiation.TOTAL_SOLAR_IRRADIANCE - radiation.SOLAR_IRRADIANCE_VARIATION)
    np.testing.assert_allclose(actual_max, expected_max, rtol=5e-5)

  def test_normalized_radiation_flux(self):
    solar_radiation = radiation.SolarRadiation.normalized(
        self.coords, self.physics_specs, self.reference_datetime)
    perihelion_datetime = datetime.datetime(1979, 1, 3)
    time = solar_radiation.datetime_to_time(perihelion_datetime)
    radiation_flux_fn = jax.jit(solar_radiation.radiation_flux)
    tisr = radiation_flux_fn(time=time)
    np.testing.assert_allclose(np.max(tisr), 1.0, rtol=5e-5)

  @parameterized.parameters(
      dict(when=datetime.datetime(1979, 1, 1, 0, 0),
           expected_time_days=0.),
      dict(when=datetime.datetime(1979, 1, 1, 0, 1),
           expected_time_days=1 / (60 * 24)),
      dict(when=datetime.datetime(1979, 1, 1, 12, 0),
           expected_time_days=0.5),
      dict(when=datetime.datetime(1979, 1, 3, 0, 0),
           expected_time_days=2.),
      dict(when=datetime.datetime(1980, 1, 1, 0, 0),
           expected_time_days=365.),
      dict(when=datetime.datetime(1981, 1, 1, 0, 0),
           expected_time_days=731.),  # 365 + 366, 1980 is a leap year
      dict(when=datetime.datetime(2020, 1, 1, 0, 0),
           expected_time_days=14975.),  # 31 * 365 + 10 * 366
      dict(when=datetime.datetime(2022, 5, 4, 4, 20),
           expected_time_days=15829.180555555555),
  )
  def test_datetime_to_time(self, when, expected_time_days):
    solar_radiation = radiation.SolarRadiation(
        self.coords, self.physics_specs, self.reference_datetime)
    actual_time = solar_radiation.datetime_to_time(when)
    expected_time = solar_radiation.physics_specs.nondimensionalize(
        expected_time_days * units.day)
    self.assertAlmostEqual(actual_time, expected_time)

  @parameterized.parameters(
      # Note, WB_REFERENCE_DATETIME = (1979, 1, 1, 0, 0)
      # Nondim time based on radiation.DAYS_PER_YEAR = 365.25
      dict(when=radiation.WB_REFERENCE_DATETIME,
           expected_orbital_phase=0,
           expected_synodic_phase=0),
      dict(when=(radiation.WB_REFERENCE_DATETIME
                 + datetime.timedelta(days=365.25)),
           expected_orbital_phase=TWOPI,
           expected_synodic_phase=365.25 * TWOPI),
      dict(when=(radiation.WB_REFERENCE_DATETIME
                 + datetime.timedelta(days=-365.25)),
           expected_orbital_phase=-TWOPI,
           expected_synodic_phase=-365.25 * TWOPI),
      dict(when=datetime.datetime(2019, 1, 1, 0, 0),
           expected_orbital_phase=40 * TWOPI,
           expected_synodic_phase=40 * 365.25 * TWOPI),
  )
  def test_time_to_orbital_time(
      self, when, expected_orbital_phase, expected_synodic_phase):
    solar_radiation = radiation.SolarRadiation(
        self.coords, self.physics_specs, self.reference_datetime)
    time = solar_radiation.datetime_to_time(when)
    actual = solar_radiation.time_to_orbital_time(time)

    expected_orbital_phase = _get_expected_value_modulo_2pi(
        expected_orbital_phase, actual.orbital_phase
    )
    self.assertAlmostEqual(actual.orbital_phase, expected_orbital_phase)

    expected_synodic_phase = _get_expected_value_modulo_2pi(
        expected_synodic_phase, actual.synodic_phase
    )
    self.assertAlmostEqual(actual.synodic_phase, expected_synodic_phase)

  def test_solar_hour_angle(self):
    solar_radiation = radiation.SolarRadiation(
        self.coords, self.physics_specs, self.reference_datetime)
    time = solar_radiation.datetime_to_time(
        datetime.datetime(1979, 4, 15, 0, 0))  # Equation of time is approx 0
    actual = solar_radiation.solar_hour_angle(time)
    lon, _ = self.coords.horizontal.nodal_mesh
    # At UTC=0, when equation of time is 0, expect hour_angle=`longitude - pi`
    expected = lon - jnp.pi
    expected = _get_expected_value_modulo_2pi(expected, actual)
    np.testing.assert_allclose(actual, expected, atol=1e-4)

  def test_nondim_minutes_day_year_constants(self):
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    days_per_year = physics_specs.nondimensionalize(units.year / units.day)
    np.testing.assert_almost_equal(days_per_year, radiation.DAYS_PER_YEAR)
    minutes_per_day = physics_specs.nondimensionalize(units.day / units.minute)
    np.testing.assert_almost_equal(minutes_per_day, radiation.MINUTES_PER_DAY)


if __name__ == '__main__':
  absltest.main()
